@@ -1,5 +1,5 @@
 // child.c (part of mintty)
-// Copyright 2008-11 Andy Koppe, 2015-2022 Thomas Wolff
+// Copyright 2008-11 Andy Koppe, 2015-2025 Thomas Wolff
 // Licensed under the terms of the GNU General Public License v3 or later.
 
 #include "child.h"
@@ -87,7 +87,7 @@ childerror(char * action, bool from_fork, int errno_code, int code)
 
   char s[33];
   bool colour_code = code && !errno_code;
-  sprintf(s, "\033[30;%dm\033[K", from_fork ? 41 : colour_code ? code : 43);
+  sprintf(s, "\r\n\033[30;%dm\033[K", from_fork ? 41 : colour_code ? code : 43);
   term_write(s, strlen(s));
   term_write(action, strlen(action));
   if (errno_code) {
@@ -159,6 +159,22 @@ open_logfile(bool toggling)
         free(log);
         log = logf;
       }
+      // also expand placeholders $h or $p with hostname or pid
+      if ((format = strstr(log, "$h"))) {
+        char hostname[HOST_NAME_MAX + 1];
+        if (0 == gethostname(hostname, HOST_NAME_MAX)) {
+          *format = '%'; *++format = 's';
+          char * logf = asform(log, hostname);
+          free(log);
+          log = logf;
+        }
+      }
+      if ((format = strstr(log, "$p"))) {
+        *format = '%'; *++format = 'd';
+        char * logf = asform(log, getpid());
+        free(log);
+        log = logf;
+      }
 
       log_fd = open(log, O_WRONLY | O_CREAT | O_EXCL, 0600);
       if (log_fd < 0) {
@@ -499,8 +515,29 @@ child_tty(void)
   return ptsname(pty_fd);
 }
 
+uchar *
+child_termios_chars(void)
+{
+static struct termios attr;
+  tcgetattr(pty_fd, &attr);
+  return attr.c_cc;
+}
+
 #define patch_319
 
+#ifdef debug_pty
+static void
+trace_line(char * tag, int n, const char * s, int len)
+{
+  printf("%s %d", tag, n);
+  for (int i = 0; i < len; i++)
+    printf(" %02X", (uint)s[i]);
+  printf("\n");
+}
+#else
+#define trace_line(tag, n, s, len)	
+#endif
+
 void
 child_proc(void)
 {
@@ -634,6 +671,7 @@ child_proc(void)
 #endif
         do {
           int ret = read(pty_fd, buf + len, sizeof buf - len);
+          trace_line("read", ret, buf + len, ret);
           //if (kb_trace) printf("[%lu] read %d\n", mtime(), ret);
           if (ret > 0)
             len += ret;
@@ -643,6 +681,8 @@ child_proc(void)
 
         if (len > 0) {
           term_write(buf, len);
+          trace_line("twrt", len, buf, len);
+
           // accelerate keyboard echo if (unechoed) keyboard input is pending
           if (kb_input) {
             kb_input = false;
@@ -719,7 +759,7 @@ static struct procinfo {
 } * ttyprocs = 0;
 static uint nttyprocs = 0;
 
-static char *
+char *
 procres(int pid, char * res)
 {
   char fbuf[99];
@@ -840,8 +880,13 @@ grandchild_process_list(void)
 void
 child_write(const char *buf, uint len)
 {
-  if (pty_fd >= 0)
+  if (pty_fd >= 0) {
+#ifdef debug_pty
+    int n =
+#endif
     write(pty_fd, buf, len);
+    trace_line("cwrt", n, buf, len);
+  }
 }
 
 /*
@@ -860,6 +905,16 @@ child_break()
   }
 }
 
+/*
+  Send an INTR char.
+ */
+void
+child_intr()
+{
+  char * c_cc = (char *)child_termios_chars();
+  child_write(&c_cc[VINTR], 1);
+}
+
 void
 child_printf(const char *fmt, ...)
 {
@@ -869,8 +924,13 @@ child_printf(const char *fmt, ...)
     char *s;
     int len = vasprintf(&s, fmt, va);
     va_end(va);
-    if (len >= 0)
+    if (len >= 0) {
+#ifdef debug_pty
+      int n =
+#endif
       write(pty_fd, s, len);
+      trace_line("tprt", n, s, len);
+    }
     free(s);
   }
 }
@@ -905,7 +965,11 @@ child_resize(struct winsize *winp)
 static int
 foreground_pid()
 {
-  return (pty_fd >= 0) ? tcgetpgrp(pty_fd) : 0;
+  int fg_pid = (pty_fd >= 0) ? tcgetpgrp(pty_fd) : 0;
+  if (fg_pid <= 0)
+    // fallback to child process
+    fg_pid = pid;
+  return fg_pid;
 }
 
 static char *
@@ -934,9 +998,13 @@ get_foreground_cwd()
 char *
 foreground_cwd()
 {
+#ifdef consider_osc7
   // if working dir is communicated interactively, use it
+  // - drop this generic handling here as it would also affect 
+  //   relative pathname resolution outside OSC 7 usage
   if (child_dir && *child_dir)
     return strdup(child_dir);
+#endif
   return get_foreground_cwd();
 }
 
@@ -1027,6 +1095,8 @@ user_command(wstring commands, int n)
   }
 }
 
+#ifdef pathname_conversion_here
+#warning now unused
 /*
    used by win_open
 */
@@ -1110,6 +1180,7 @@ child_conv_path(wstring wpath, bool adjust_dir)
 
   return win_wpath;
 }
+#endif
 
 void
 child_set_fork_dir(char * dir)
@@ -1143,6 +1214,7 @@ setup_sync(bool in_tabs)
     }
     if (cfg.tabbar) {
       setenvi("MINTTY_TABBAR", cfg.tabbar);
+      setenvi("MINTTY_SYNC", cfg.geom_sync);
       // enforce proper grouping, i.e. either new tab or window, as requested
       if (in_tabs && *cfg.class) {
         char * class = cs__wcstoutf(cfg.class);
@@ -1196,8 +1268,14 @@ do_child_fork(int argc, char *argv[], int moni, bool launch, bool config_size, b
 
   if (clone == 0) {  // prepare child process to spawn new terminal
     string set_dir = 0;
-    if (in_cwd)
-      set_dir = get_foreground_cwd(false);  // do this before close(pty_fd)!
+    if (in_cwd) {
+      if (support_wsl) {
+        if (child_dir && *child_dir)
+          set_dir = strdup(child_dir);
+      }
+      else
+        set_dir = get_foreground_cwd();  // do this before close(pty_fd)!
+    }
 
     if (pty_fd >= 0)
       close(pty_fd);
@@ -1209,6 +1287,8 @@ do_child_fork(int argc, char *argv[], int moni, bool launch, bool config_size, b
       if (set_dir) {
         // use cwd of foreground process if requested via in_cwd
       }
+#ifdef pathname_conversion_here
+#warning now deprecated; handled via guardpath
       else if (support_wsl) {
         wchar * wcd = cs__utftowcs(child_dir);
 #ifdef debug_wsl
@@ -1221,8 +1301,14 @@ do_child_fork(int argc, char *argv[], int moni, bool launch, bool config_size, b
         set_dir = (string)cs__wcstombs(wcd);
         delete(wcd);
       }
-      else
-        set_dir = strdup(child_dir);
+#endif
+      else {
+        //set_dir = strdup(child_dir);
+        set_dir = guardpath(child_dir, 2);
+        if (!set_dir)
+          sleep(1);  // let beep be audible before execv below
+          // also skip chdir below (guarded)
+      }
 
       if (set_dir) {
         chdir(set_dir);
@@ -1342,6 +1428,8 @@ child_fork(int argc, char *argv[], int moni, bool config_size, bool in_cwd, bool
 void
 child_launch(int n, int argc, char * argv[], int moni)
 {
+  setup_sync(true);
+
   if (*cfg.session_commands) {
     char * cmds = cs__wcstombs(cfg.session_commands);
     char * cmdp = cmds;
