@@ -169,12 +169,21 @@ enum {
   TATTR_CLEAR     = 0x4000000000000000u, /* erased / unwritten */
   TATTR_OVERHANG  = 0x0080000000000000u, /* visual double-width overhang */
 
+  TATTR_JOINED    = 0x0000080000000000u, /* LAM/ALEF joined as combining */
+
   DATTR_STARTRUN  = 0x8000000000000000u, /* start of redraw run */
   DATTR_MASK      = TATTR_RIGHTCURS | TATTR_PASCURS | TATTR_ACTCURS
-                    | DATTR_STARTRUN
+                    | DATTR_STARTRUN,
+
+#ifdef configured_glyph_shift
+  GLYPHSHIFT_MASK = 0x0000070000000000u,
+  ATTR_GLYPHSHIFT_SHIFT = 40,
+#define GLYPHSHIFT_MAX (GLYPHSHIFT_MASK >> ATTR_GLYPHSHIFT_SHIFT)
+#else
+  ATTR_GLYPHSHIFT = 0x0000040000000000u, /* alternative font glyph shift */
+#endif
+
   // unassigned bits:
-  //              = 0x0000080000000000u,
-  //              = 0x0000040000000000u,
   //              = 0x0000020000000000u,
   //              = 0x0000010000000000u,
 };
@@ -398,6 +407,8 @@ typedef struct imglist {
   // position within scrollback
   long long int top;  // includes offset term.virtuallines
   int left;
+  // last painted position on screen, to avoid repetitive painting
+  short x, y;
 
   // image area (cell units)
   int width, height;
@@ -561,6 +572,7 @@ struct term {
   bool wide_indic;
   bool wide_extra;
   bool disable_bidi;
+  bool join_lam_alef;
   bool enable_bold_colour;
   bool enable_blink_colour;
   bool readline_mouse_1;
@@ -658,6 +670,7 @@ struct term {
   int locator_top, locator_left, locator_bottom, locator_right;
 
   bool selected, sel_rect;
+  bool selection_eq_clipboard;
   pos sel_start, sel_end, sel_anchor;
   bool hovering;
   int hoverlink;
