@@ -1,5 +1,5 @@
 // termclip.c (part of mintty)
-// Copyright 2008-10 Andy Koppe, 2018 Thomas Wolff
+// Copyright 2008-23 Andy Koppe, 2024-2025 Thomas Wolff
 // Adapted from code from PuTTY-0.60 by Simon Tatham and team.
 // Licensed under the terms of the GNU General Public License v3 or later.
 
@@ -20,10 +20,11 @@ typedef struct {
 } clip_workbuf;
 
 static void
-destroy_clip_workbuf(clip_workbuf * b)
+destroy_clip_workbuf(clip_workbuf * b, bool with_text)
 {
   assert(b && b->capacity); // we're only called after get_selection, which always allocates
-  free(b->text);
+  if (with_text)
+    free(b->text);
   if (b->with_attrs)
     // the attributes part of the buffer was only filled as requested
     free(b->cattrs);
@@ -78,8 +79,10 @@ clip_addchar(clip_workbuf * b, wchar chr, cattr * ca, bool tabs, ulong sizehint)
   }
 
   cattr copattr = ca ? *ca : CATTR_DEFAULT;
+  //printf("setting clipbuf[%ld] = %02X\n", b->len, chr);
   if (copattr.attr & TATTR_CLEAR) {
-    copattr.attr &= ~(ATTR_BOLD | ATTR_DIM | TATTR_CLEAR);
+    if (!tabs)
+      copattr.attr &= ~(ATTR_BOLD | ATTR_DIM | TATTR_CLEAR);
   }
 
   b->text[b->len] = chr;
@@ -96,6 +99,8 @@ get_selection(bool attrs, pos start, pos end, bool rect, bool allinline, bool wi
 {
   //printf("get_selection attrs %d all %d tabs %d\n", attrs, allinline, with_tabs);
 
+  if (with_tabs)
+    attrs = true;  // ensure we can check expanded TABs (#1269)
   clip_workbuf *buf = newn(clip_workbuf, 1);
   *buf = (clip_workbuf){.with_attrs = attrs,
                         .capacity = 0, .len = 0, .text = 0, .cattrs = 0};
@@ -216,7 +221,10 @@ get_selection(bool attrs, pos start, pos end, bool rect, bool allinline, bool wi
     }
     if (nl) {
       clip_addchar(buf, '\r', 0, false, hint);
-      clip_addchar(buf, '\n', 0, false, hint);
+      // mark lineend with line attributes, particularly double-width/height
+      cattr lcattr = CATTR_DEFAULT;
+      lcattr.link = line->lattr;
+      clip_addchar(buf, '\n', &lcattr, false, hint);
     }
     start.y++;
     start.x = rect ? old_top_x : 0;
@@ -233,7 +241,7 @@ get_sel_str(pos start, pos end, bool rect, bool allinline, bool with_tabs)
 {
   clip_workbuf * buf = get_selection(false, start, end, rect, allinline, with_tabs);
   wchar * selstr = buf->text;
-  free(buf);
+  destroy_clip_workbuf(buf, false);
   return selstr;
 }
 
@@ -251,7 +259,7 @@ term_copy_as(char what)
   // for CopyAsHTML, get_selection will be called another time
   // but with different parameters
   win_copy_as(buf->text, buf->cattrs, buf->len, what);
-  destroy_clip_workbuf(buf);
+  destroy_clip_workbuf(buf, true);
 }
 
 void
@@ -272,44 +280,109 @@ term_open(void)
   wchar * p = selstr;
   while (iswspace(*p))
     p++;
-  if (*p) {
-    wchar * url = p;
-    while (*p && !iswspace(*p))
-      p++;
-    *p = 0;
-    win_open(wcsdup(url), true);  // win_open frees its argument
+
+  if (*p)
+    win_open(selstr, true);  // win_open frees its argument
+  else
+    free(selstr);
+}
+
+static bool filter_NUL;
+static char filter[69];
+
+static void
+set_filter(string s)
+{
+  bool do_filter(string tag)
+  {
+#if CYGWIN_VERSION_API_MINOR >= 171
+    char * match = strcasestr(s, tag);
+#else
+    char * match = strstr(s, tag);
+#endif
+    //return match;  // a bit simplistic, we should probably properly parse...
+    if (!match)
+      return false;
+    return (match == s || *(match - 1) < '@')
+           && match[strlen(tag)] < '@';
   }
 
-  free(selstr);
+  filter_NUL = false;
+  *filter = 0;
+  if (do_filter("C0")) {
+    filter_NUL = true;
+    strcat(filter, "\t\n\r");
+  }
+  else {
+    if (do_filter("BS")) strcat(filter, "\b");
+    if (do_filter("HT")) strcat(filter, "\t");
+    if (do_filter("NL")) strcat(filter, "\n");
+    if (do_filter("CR")) strcat(filter, "\r");
+    if (do_filter("FF")) strcat(filter, "\f");
+    if (do_filter("ESC")) strcat(filter, "\e");
+  }
+  if (do_filter("DEL"))
+    strcat(filter, "\177");
+  if (do_filter("C1")) {
+    filter_NUL = true;
+    strcat(filter, "\200\201\202\203\204\205\206\207"
+                   "\210\211\212\213\214\215\216\217"
+                   "\220\221\222\223\224\225\226\227"
+                   "\230\231\232\233\234\235\236\237");
+  }
+  if (do_filter("STTY")) {
+    int i = strlen(filter);
+    void addchar(wchar c)
+    {
+      if (c)
+        filter[i++] = c;
+      else
+        filter_NUL = true;
+    }
+    uchar * c_cc = child_termios_chars();
+    addchar(c_cc[VINTR]);
+    addchar(c_cc[VQUIT]);
+    addchar(c_cc[VSUSP]);
+    addchar(c_cc[VSWTC]);
+    filter[i] = 0;
+  }
 }
 
 static bool
-contains(string s, wchar c)
+isin_filter(wchar c)
 {
-  string tag;
-  switch (c) {
-    when '\b': tag = "BS";
-    when '\t': tag = "HT";
-    when '\n': tag = "NL";
-    when '\r': tag = "CR";
-    when '\f': tag = "FF";
-    when '\e': tag = "ESC";
-    when '\177': tag = "DEL";
-    otherwise:
-      if (c < ' ')
-        tag = "C0";
-      else if (c >= 0x80 && c < 0xA0)
-        tag = "C1";
-      else
-        return false;
-  }
-  return strstr(s, tag);
-  // a bit simplistic, we should probably properly parse...
+  if (c & 0xFFFFFF00)
+    return false;
+  if (!c)
+    return filter_NUL;
+  return strchr(filter, c);
 }
 
 void
 term_paste(wchar *data, uint len, bool all)
 {
+  // set/refresh list of characters to be filtered;
+  // stty settings may have changed
+  set_filter(cfg.filter_paste);
+
+  if (cfg.confirm_multi_line_pasting
+      && !(strchr(filter, '\r') && strchr(filter, '\n')))
+  {
+    // check multi-line pasting
+    bool multi_line = false;
+    for (uint i = 0; i < len; i++) {
+      if (data[i] == '\r' || data[i] == '\n') {
+        multi_line = true;
+        break;
+      }
+    }
+    if (multi_line
+        && !win_confirm_text(data, W("Multi-line pasting – confirm?")))
+    {
+      return;
+    }
+  }
+
   term_cancel_paste();
 
   uint size = len;
@@ -335,7 +408,7 @@ term_paste(wchar *data, uint len, bool all)
     wchar wc = data[i];
     if (wc == '\n')
       wc = '\r';
-    if (!all && *cfg.filter_paste && contains(cfg.filter_paste, wc))
+    if (!all && *cfg.filter_paste && isin_filter(wc))
       wc = ' ';
 
     if (data[i] != '\n')
@@ -411,6 +484,7 @@ term_select_all(void)
   term.sel_start = (pos){-sblines(), 0, 0, 0, false};
   term.sel_end = (pos){term_last_nonempty_line(), term.cols, 0, 0, true};
   term.selected = true;
+  term.selection_eq_clipboard = false;
   if (cfg.copy_on_select)
     term_copy();
 }
@@ -635,6 +709,12 @@ term_create_html(FILE * hf, int level)
     );
   if (level >= 3)
     hprintf(hf, "  body.mintty { margin: 0; padding: 0; }\n");
+  hprintf(hf, "  .super, .sub, .small { line-height: 0; font-size: 0.7em; letter-spacing: 0.3em; }\n");
+  hprintf(hf, "  .super { vertical-align: super; }\n");
+  hprintf(hf, "  .sub { vertical-align: sub; }\n");
+  hprintf(hf, "  .double-width {display: inline-block; line-height: 1.2; transform-origin: left; transform: scale(2, 1);}\n");
+  hprintf(hf, "  .double-height-top {display: inline-block; line-height: 2.4; transform-origin: left; transform: scale(2, 2);}\n");
+  hprintf(hf, "  .double-height-bottom {display: none;}\n");
   hprintf(hf, "  #vt100 span {\n");
   if (level >= 2) {
     // font needed in <span> for some tools (e.g. Powerpoint)
@@ -686,7 +766,23 @@ term_create_html(FILE * hf, int level)
         hprintf(hf, "  .background {\n");
       }
 
-      hprintf(hf, "    background-image: url('%s');\n", bg);
+      char * bgg = guardpath(bg, 4);
+      if (bgg) {
+        wchar * bgw = path_posix_to_win_w(bgg);
+        free(bgg);
+        free(bg);
+        bg = cs__wcstoutf(bgw);
+        free(bgw);
+        char * cc = bg;
+        while (*cc) {
+          if (*cc == '\\')
+            *cc = '/';
+          cc++;
+        }
+
+        hprintf(hf, "    background-image: url('file:///%s');\n", bg);
+      }
+
       if (!tiled) {
         hprintf(hf, "    background-attachment: no-repeat;\n");
         hprintf(hf, "    background-size: 100%% 100%%;\n");
@@ -775,11 +871,13 @@ term_create_html(FILE * hf, int level)
   hprintf(hf, "<body class=mintty onload='setup();'>\n");
   //hprintf(hf, "  <table border=0 cellpadding=0 cellspacing=0><tr><td>\n");
   hprintf(hf, "  <div class=background id='vt100'>\n");
-  hprintf(hf, "   <pre>");
+  hprintf(hf, "   <pre\n>");
 
   clip_workbuf * buf = get_selection(true, start, end, rect, level >= 3, false);
   int i0 = 0;
   bool odd = true;
+  bool new_line = true;
+  ushort lattr = LATTR_NORM;
   for (uint i = 0; i < buf->len; i++) {
     if (!buf->text[i] || buf->text[i] == '\r' || buf->text[i] == '\n'
         // buf->cattrs[i] ~!= buf->cattrs[i0] ?
@@ -792,6 +890,21 @@ term_create_html(FILE * hf, int level)
         || buf->cattrs[i].ulcolr != buf->cattrs[i0].ulcolr
        )
     {
+      if (new_line) {
+        wchar * nl = wcschr(&buf->text[i], '\n');
+        if (nl) {
+          int offset = nl - &buf->text[i];
+          lattr = (ushort)buf->cattrs[i + offset].link & LATTR_MODE;
+        }
+        else
+          lattr = LATTR_NORM;
+        if (lattr)
+          hprintf(hf, "<div class='double-%s'>",
+                      lattr == LATTR_WIDE ? "width" :
+                      lattr == LATTR_TOP ? "height-top" : "height-bottom");
+        new_line = false;
+      }
+
       // flush chunk with equal attributes
       hprintf(hf, "<span class='%s", odd ? "od" : "ev");
 
@@ -838,6 +951,14 @@ term_create_html(FILE * hf, int level)
       if (ca->attr & ATTR_FRAMED)
         hprintf(hf, " emoji");  // mark emoji style
 
+      // add subscript or superscript
+      if ((ca->attr & (ATTR_SUBSCR | ATTR_SUPERSCR)) == (ATTR_SUBSCR | ATTR_SUPERSCR))
+        hprintf(hf, " small");
+      else if (ca->attr & ATTR_SUBSCR)
+        hprintf(hf, " sub");
+      else if (ca->attr & ATTR_SUPERSCR)
+        hprintf(hf, " super");
+
       // style adding function
       bool with_style = false;
       void add_style(char * s) {
@@ -1023,16 +1144,39 @@ term_create_html(FILE * hf, int level)
         // more precise cursor colour adjustments could be made...
       }
 
+      // finish styles
+      hprintf(hf, "'>");
+
       // retrieve chunk of text from buffer
       wchar save = buf->text[i];
       buf->text[i] = 0;
       char * s = cs__wcstoutf(&buf->text[i0]);
       buf->text[i] = save;
+      // here we could:
+      // * handle the chunk string by Unicode glyphs
+      // * check whether each char is an emoji char or sequence
+      // * check its terminal width
+      // * scale width to actual (narrow or multi-cell) width
 
       // write chunk, apply HTML escapes
+      void hprinttext(char * t) {
+        if (ca->attr & ATTR_FRAMED)
+          while (*t) {
+            hprintf(hf, "%c", *t++);
+#ifdef export_emoji_style
+            // here we should, in addition to the above:
+            // * check whether each char actually has an emoji presentation:
+            //   (emoji_tags(emoji_idx(ch)) & EM_emoj)
+            //   and only append 0xFE0F then
+            if ((*t & 0xC0) != 0x80)
+              hprintf(hf, "️");
+#endif
+          }
+        else
+          hprintf(hf, "%s", t);
+      }
       char * s1 = strpbrk(s, "<&");
       if (s1) {
-        hprintf(hf, "'>");
         char * s0 = s;
         do {
           if (*s0 == '<') {
@@ -1047,7 +1191,7 @@ term_create_html(FILE * hf, int level)
             char c = s1 ? *s1 : 0;
             if (s1)
               *s1 = 0;
-            hprintf(hf, "%s", s0);
+            hprinttext(s0);
             if (s1) {
               *s1 = c;
               s0 = s1;
@@ -1057,11 +1201,11 @@ term_create_html(FILE * hf, int level)
           }
           s1 = strpbrk(s0, "<&");
         } while (*s0);
-        hprintf(hf, "</span>");
       }
       else
-        hprintf(hf, "'>%s</span>", s);
+        hprinttext(s);
       free(s);
+      hprintf(hf, "</span>");
 
       // forward chunk pointer
       i0 = i;
@@ -1075,15 +1219,20 @@ term_create_html(FILE * hf, int level)
     if (buf->text[i] == '\n') {
       i++;
       i0 = i;
+      if (lattr)
+        hprintf(hf, "</div>");
       if (enhtml)
-        // <br> needed for Powerpoint
-        hprintf(hf, "<br\n>");
+        // <br> needed for HTML and for Powerpoint
+        hprintf(hf, "<br%s\n>", lattr == LATTR_BOT ? " class='double-height-bottom'" : "");
       else
         hprintf(hf, "\n");
       odd = !odd;
+
+      new_line = true;
+      lattr = LATTR_NORM;
     }
   }
-  destroy_clip_workbuf(buf);
+  destroy_clip_workbuf(buf, true);
 
   hprintf(hf, "</pre>\n");
   hprintf(hf, "  </div>\n");
@@ -1146,6 +1295,6 @@ print_screen(void)
   clip_workbuf * buf = get_selection(false, start, end, rect, false, false);
   printer_wwrite(buf->text, buf->len);
   printer_finish_job();
-  destroy_clip_workbuf(buf);
+  destroy_clip_workbuf(buf, true);
 }
 
